'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-09 16:55:08
FilePath: /outlier/src/outlier/tools/utils.py
'''
import logging
from logging import handlers
import socket
import sys

FORMAT = '%(asctime)s - %(message)s'



def initlogger(level, format=FORMAT, filehandlename=None):
    """To init log with certain level and format,
    or designate a filename for rotating file log
    handler

    Args:
        level (_type_): level
        format (_type_, optional): _description_. Defaults to FORMAT as `%(asctime)s - %(message)s`.
        filehandlename (_type_, optional): _description_. Defaults to None.
    """
    
    if filehandlename is not None:
        logging.basicConfig(format=format, level=eval("logging."+level), handlers=[handlers.RotatingFileHandler(filehandlename, maxBytes=500000, backupCount=20)])
    else:
        logging.basicConfig(format=format, level=eval("logging."+level))
    
    
 
def gethostaddr() -> str:
    """This function is to check the ipaddress that a host should be bind

    Returns:
        str: ip address
    """
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()
        return ip[0]


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-09 11:54:11
FilePath: /outlier/src/outlier/biz/beans.py
'''
from dataclasses import dataclass
from datetime import datetime
from typing import List


@dataclass
class ChatMessage:
    name:       str
    ts:         float
    content:    str

    def getattrs(self):
        return {'name': self.name, 'timestamp': self.ts, 'content': self.content}
    
    @staticmethod
    def parse(cm: dict):
        return ChatMessage(cm['name'], cm['timestamp'], cm['content'])
    
    @staticmethod
    def get_round_time(timestamp: float) -> str:
        """This is to convert timestamp gap to a human-readable time gap
        such as `1 days ago`, `1 hours ago` etc.

        Args:
            timestamp (float): target timestamp

        Returns:
            str: time gap
        """
        curr = datetime.now()
        happening = datetime.fromtimestamp(timestamp)
        delta = curr - happening
        if (delta.days > 0):
            return f"{delta.days} days ago"
        elif delta.seconds // 3600 > 0:
            return f"{delta.seconds // 3600} hours ago"
        elif delta.seconds // 60 > 0:
            return f"{delta.seconds // 60} minutes ago"
        else:
            return "just now"
    
    def format(self):
        return f"{self.name} -- {ChatMessage.get_round_time(self.ts)}: {self.content}\n"
    

class Room:
    index = 0
    def __init__(self, name) -> None:
        self.history: List[ChatMessage] = list()
        self.group:   int               = Room.index
        self.passwd:  str               = None
        self.name:    str               = name
        self.connects:int               = 0
        self.lastact: float             = -1
        Room.index += 1
        
    
    def __repr__(self) -> str:
        l = ("PSWD" if self.passwd else "FREE")
        return f"{'room:'+str(self.name):10}\t{l}\t{self.connects}\t{len(self.history)}"
    
    @staticmethod
    def header():
        return f"{'name':10}\t{'LOCK'}\tpeople\tmsg"
    


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-09 16:55:14
FilePath: /outlier/src/outlier/tools/decorators.py
'''
import time
from typing import Callable
import os
import logging

def onexit(func: Callable):
    def wrapped_func(*args, **kwargs):
        logging.debug(f"Exiting calling {func}")
        func(*args, **kwargs)
        time.sleep(0.1)
        print("Exited Bye")
        os._exit(0)
    return wrapped_func


def singleton(clazz):
    clazz.ins = None
    clazz.inited = False
    clazz.origin_init = clazz.__init__
    
    def wrappednew(cls, *args, **kwargs):
        if clazz.ins is None:
            # print(f"created cls {cls}")
            clazz.ins = super(clazz, cls).__new__(cls)
        return clazz.ins
    
    def wrappedinit(self, *args, **kwargs):
        if not clazz.inited:
            # print(f"init cls {self}")
            clazz.origin_init(self, *args, **kwargs)
            clazz.inited = True
            
    
    clazz.__new__ = wrappednew
    clazz.__init__ = wrappedinit
    return clazz



'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-09 16:56:01
FilePath: /outlier/src/outlier/tools/threadpool.py
'''
import queue
import threading
import time
import logging
from typing import Callable
from traceback import print_exc
from tools.decorators import singleton

class WorkThread(threading.Thread):
    def __init__(self, target):
        super(WorkThread, self).__init__(target=target)
        self.setDaemon(True)
        self.start()
        
@singleton
class ThreadPool:
    def __init__(self, max_size=10, show_exception=True, status_monitor=False) -> None:
        self.queue = queue.Queue()
        self.waiting_task_num = 0
        self.processing_task_num = 0
        self.status_monitor = status_monitor
        self.show_exception = show_exception
        self.threads = [WorkThread(target=self._work) for _ in range(max_size)]
    
    def _work(self):
        while True:
            task, args = self.queue.get()
            try:
                self.waiting_task_num -= 1
                self.processing_task_num += 1
                task(*args)
            except Exception as e:
                logging.debug(f"warning some work failed {e}")
                if self.show_exception:
                    print_exc()
                time.sleep(1)
            finally:
                self.processing_task_num -= 1
                self.queue.task_done()
        
    def put_task(self, task: Callable, args=None):
        self.waiting_task_num += 1
        for thread in self.threads:
            if not thread.is_alive():
                logging.debug(f"Some thread dead remove it {thread}")
                self.threads.remove(thread)
                logging.debug(f"Add a new thread")
                self.threads.append(WorkThread(target=self._work))
    
        self.queue.put((task, args if args is not None else ()))
        
    def close(self):
        while not self.queue.empty():
            self.queue.get_nowait()
            
        
    def check_status(self):
        logging.debug(f"Working thread {len(self.threads)} waiting task {self.waiting_task_num}, processing task {self.processing_task_num}")
        
    # def open_status_monitor(self):
    #     self.status_monitor = True
    #     self.put_task(self.check_status_loop)
        
    # def close_status_monitor(self):
    #     self.status_monitor = False
        
    # def check_status_loop(self):
    #     import time
    #     while self.status_monitor:
    #         self.check_status()
    #         time.sleep(0.5)
        
    @staticmethod
    def template_loop(r):
        import time
        for i in range(r):
            logging.debug(f"[TASK]{r} counts at {i}/{r}")
            time.sleep(1)
            
# tp = ThreadPool(max_size=3)
# tp.put_task(ThreadPool.template_loop, args=(1,))
# tp.put_task(ThreadPool.template_loop, args=(2,))
# tp.put_task(ThreadPool.template_loop, args=(3,))
# tp.put_task(ThreadPool.template_loop, args=(4,))

# input()


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-10 02:43:33
FilePath: /outlier/src/outlier/biz/bizservice.py

This is the key layer for bussniness implementation. We use server 
as the `stem` update way, client follows. we use some way to add the biz
respond function automatically and we use some way to make the client can
easily correspoing to it.
For exmaple,

`python`

def bizserv:

class Server:
    @bizserv
    def bizfuncA:
    
    @bizserv
    def bizfuncB:

def bizclnt

class Client:
    @bizclnt(bizfuncA)
    def bizfuncA:
    
    @bizclnt(bizfunc)
    def bizfuncB:


We have three phase in total.
1. generate request -> biz and this request can be related with biz func
2. request -> package
3. package -> request 
4. generate response -> biz
5. repsonse -> package
6. solve repsonse -> biz
'''
import inspect
import logging
import re
import signal
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, List

from ..encryption.sessionservice import Package, Session, SessionService
from ..transmission.tcpservice import Ops


class State(Enum):
    IDLE = -1
    Hall = 0
    Room = 1
    Cfrm = 2
    
    

@dataclass
class User:
    sess: Session
    name: str
    syncts: float = -1
    
@dataclass
class BizRequest:
    cmd: str   = None
    ts:  float = None
    param: Any = None

@dataclass
class BizResponse:
    resp: Any   = None
    bcresp: Any = None
    ts: float   = None
    cmd: str    = None
    group: int  = -1
    inc: bool   = False

class BizService:
    def __init__(self, sessservice: SessionService, **kwargs) -> None:
        self.sessservice = sessservice
        self.sessservice.set_upper_rchandle(self.rchandle)
        self.acfuncs: List[Callable] = list()
        self.rcfuncs: List[Callable] = list()
        self.kwargs = kwargs
        signal.signal(signal.SIGINT, self.close)
        
    def rchandle(self, ops: Ops, session: Session, package: Package, *args):
        pass
        
    def _pack_req(self, bizreq: BizRequest) -> Package:
        return Package.buildpackage().add_cmd(bizreq.cmd) \
            .add_field_if(bizreq.param is not None, "param", bizreq.param) \
            .add_field_if(bizreq.ts is not None, Package.TIME, bizreq.ts)
        
    def _unpack_req(self, package: Package) -> BizRequest:
        return BizRequest(package.get_field("cmd"), package.get_field("ts"), package.get_field("param"))
    
    def _pack_resp(self, bizresp: BizResponse) -> Package:
        pack = Package.buildpackage().add_cmd(bizresp.cmd) \
            .add_field("param", bizresp.resp) \
            .add_field_if(bizresp.ts is not None, Package.TIME, bizresp.ts)
            
        bc = None if bizresp.bcresp is None else Package.buildpackage().add_cmd(bizresp.cmd) \
            .add_field("param", bizresp.bcresp) \
            .add_field_if(bizresp.ts is not None, Package.TIME, bizresp.ts)
        return pack, bc
    
    
    def _getrcfuncs(self, ops: Ops, bizreq: BizRequest) -> Callable:
        for func in self.rcfuncs:
            if func.__qualname__.partition('.')[-1] == bizreq.cmd:
                return func
            if 'rc' in dir(func) and func.rc == ops:
                return func
        return lambda *args:  BizResponse("NA")
    
    def close(self, *args):
        self.sessservice.close(*args)
    
    def start(self):
        self.sessservice.start()


def bizserv(**kwargs):
    """This decorator can make a method in server to be a busniess function
    and automatically do works around the method list as follows:
    1. Automatically registered in the server and handle the comming message
    if command inside corresponds.
    2. Send the response to client for the requesting client and broadcast the message
    for the group designated.
    3. Once `rc` is set, it can handle the corresponding Ops, if required.
    
    """
    def inner(fn: Callable[[BizService, User, BizRequest, Any], BizResponse]):
        def wrapper(bizservice, user: User, bizreq: BizRequest, *args, **kwargs):
            bizresp: BizResponse = fn(bizservice, user, bizreq, *args, **kwargs)
            bizresp.cmd = fn.__qualname__.partition('.')[-1]
            bizservice.send(user, bizresp)
            return bizresp
        wrapper.__qualname__ = fn.__qualname__
        wrapper.__setattr__("wrapped", "server")
        wrapper.__setattr__("kwargs", kwargs)
        
        if kwargs is not None and 'rc' in kwargs:
            wrapper.__setattr__('rc', kwargs.get('rc'))
        return wrapper
    return inner



class ServerBizService(BizService):
    def __init__(self, sessservice: SessionService, **kwargs) -> None:
        super().__init__(sessservice, **kwargs)
        self.users: Dict[Any, User] = dict()
        for name, func in inspect.getmembers(self, predicate=inspect.ismethod):
            if 'wrapped' in dir(func):
                logging.debug(f"{name} registered on {func}")
                self.rcfuncs.append(func)
        
    
    def rchandle(self, ops: Ops, session: Session, package: Package, *args):
        logging.debug(f"[Biz Layer] recall {ops} {session.conn.addr} {package}")
        bizreq = self._unpack_req(package)
        user = self.users.get(session.conn.addr)
        if ops == Ops.Add:
            user = User(session, "default")
            self.users[session.conn.addr] = user
        elif ops == Ops.Rmv:
            self.users.pop(session.conn.addr)
        elif ops == Ops.Rcv:
            pass
        
        bizfunc = self._getrcfuncs(ops, bizreq)
        bizfunc(user, bizreq)
        
        
        
    def send(self, user:User, bizresp: BizResponse):
        pack, bc = self._pack_resp(bizresp)
        if not bizresp.inc:
            self.sessservice.send(pack, user.sess)
        if bc:
            bcusers = filter(lambda u: u.sess.group == bizresp.group, self.users.values())
            bcusers = filter(lambda u: u != user if not bizresp.inc else True, bcusers)
            bcsess = map(lambda u: u.sess, bcusers)
            self.sessservice.send_group(bc, *bcsess)
        
        

        
        

    
def bizclnt(state:State, 
            invoke:str = None, 
            invokeptn: str = None,  
            bindto: Callable = None, 
            recall: Callable = None, 
            **kwargs):
    """This decorator can make a method in client to be a busniess function
    and automatically do works around the method list as follows:
    
    1. Automatically registered in the client.
    2. Send the request to the server
    3. Designate a state can be invoked
    4. Designate a invoke command or pattern
    5. Designate a responding method of the server
    6. Designate a recall method once the message comes from the server
    

    Args:
        state (State): _description_
        invoke (str, optional): _description_. Defaults to None.
        invokeptn (str, optional): _description_. Defaults to None.
        bindto (Callable, optional): _description_. Defaults to None.
        recall (Callable, optional): _description_. Defaults to None.
    """
    def gethelp():
        usage =  kwargs.get("usage", None)
        if invoke is not None:
            usage =  (invoke + ((" "+kwargs.get("descparams", "")) if "descparams" in kwargs else "")) if usage is None else usage
            return "\t"+ invoke +"\t" + kwargs.get("desc", "") + "\n\t\t-Usage:\t" + usage
        else:
            return "\t"+ f"hidden method {invoke} {invokeptn} {bindto} {recall}. Do not invoke" 
        
    
    def inner(fn: Callable[[BizService, str, str, Any], BizRequest]):
        def wrapper(bizservice: BizService, inputs: str = None, *args, **kwargs):
            try:
                bizreq: BizRequest = fn(bizservice, inputs, *args, **kwargs)
                if bindto is not None:
                    bizreq.cmd = bindto.__qualname__.partition(".")[-1]
                bizservice.send(bizreq)
                return bizreq
            except Exception as e:
                print(e)
                print(wrapper.help)
                return BizRequest()
       

        # set basic attribute: [name, wrapped, state]        
        wrapper.__qualname__ = fn.__qualname__
        wrapper.__setattr__("wrapped", "clnt")
        wrapper.__setattr__("state", state)
        wrapper.__setattr__("help", gethelp())
        
        # set invoke string to make sure whether it should be invoked by certain command
        if invoke is not None:
            wrapper.__setattr__("invokestr", invoke)
            
        # set invoke command pattern
        if invoke is not None or invokeptn is not None:
            wrapper.__setattr__("invoke", lambda cmd, atstate: \
                (atstate  == state and (invoke == cmd if invokeptn is None else re.match(invokeptn, cmd) is not None)))
            
        # set server proc method
        if bindto is not None:
            # logging.debug(f"{fn} binds to {bindto.__qualname__}")
            wrapper.__setattr__("bindtoname", bindto.__qualname__.partition(".")[-1])
            
        # set recall method to proc server response
        if recall is not None:
            # logging.debug(f"{fn} recall at {recall.__qualname__}")
            wrapper.__setattr__("recall", recall)
            recall.__setattr__("recall", "clnt")
            
        return wrapper
    return inner
    
    
        
class ClientBizService(BizService):
    def __init__(self, sessservice: SessionService, **kwargs) -> None:
        super().__init__(sessservice, **kwargs)
        for name, func in inspect.getmembers(self, predicate=inspect.ismethod):
            if 'wrapped' in dir(func):
                logging.debug(f"{name} registered on {func}")
                self.acfuncs.append(func)
        self.atstate = State.IDLE
        
    def rchandle(self, ops: Ops, session: Session, package: Package, *args):
        bizreq = self._unpack_req(package)
        for func in self.acfuncs:
            if 'bindtoname' in dir(func) and  func.bindtoname == bizreq.cmd and 'recall' in dir(func):
                recallfunc = func.recall
                recallfunc(self, package)
        
    def send(self, bizreq: BizRequest):
        self.sessservice.send(self._pack_req(bizreq))
        
    
    
    def process_input(self, inputs: str):
        if len(inputs.strip()) <= 0: return
        splits = inputs.split()
        cmd, params = splits[0], splits[1:]
        
        searchfunc = None
        for func in self.acfuncs:
            if 'invoke' in dir(func) and func.invoke(cmd, self.atstate):
                searchfunc = func
                
        if searchfunc is None:
            print("Command failed")
            self.gethelp(inputs, *params)
        else:
            searchfunc(inputs, *params)
            
    
    def gethelp(self, inputs=None, *args, **kwargs):
        print(f"Current at {self.atstate.name} - Supported commands:\n")
        
        for func in self.acfuncs:
            if func.state == self.atstate and 'invokestr' in dir(func):
                print(f"{func.help}\n")
                
        




            
        
    
    
    
    

        


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-10 02:47:10
FilePath: /outlier/src/outlier/transmission/tcpservice.py

This module is to provide stable and reliable layer communication as tcp protocol
'''

import logging
import select
import socket
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Dict

from ..tools.aux import Ops
from ..tools.decorators import onexit, singleton
from ..tools.threadpool import ThreadPool


@dataclass
class Connection:
    sock: socket.socket
    addr: Any
    last: float
    
    
class TcpService:
    
    """This service is to use epoll to provide a bi-direction
    tcp channel, we hope it can be general for both the server and
    the client
    """
    def __init__(self, is_block: bool = False, **kwargs) -> None:
        self.sock = socket.socket()
        self.epctl = select.epoll()
        self.threadpool = ThreadPool()           
        self.is_block = is_block
        self.kwargs = kwargs
        self.loop = True
        self.upper_rchandle: Callable[[Ops, Connection, bytes, Any], Any] = lambda *args: ...
        self.conn = Connection(self.sock, None, -1)
            
    def send(self, byteflow: bytes, conn: Connection = None):
        """Actively send message to connection

        Args:
            byteflow (bytes): data
            conn (Connection, optional): connection.
        """
        if conn:    
            conn.sock.send(byteflow)
            
        
    def set_upper_rchandle(self, upper_rchandle: Callable[[Ops, Connection, bytes, Any], Any]):
        """This is provided for upper service to invoke recall funcion.
        Invoked at the time when socket received buffer data.
        Args:
            upper_rchandle (Callable[[Ops, Connection, bytes, Any], Any]): The func to be invoked by upper layer.
        """
        self.upper_rchandle = upper_rchandle
    
    def _rchandle(self, ops: Ops, conn: Connection, fd:int = -1, byteflow: bytes = None, *args):
        """Handling the event when something happens on the socket, for server, the event
        for ADD, RECEIVE and REMOVE should be handled, and for clients, the event of REVEIVE and
        ERROR connection should be handled

        Args:
            ops (Ops): Ops
            conn (Connection): connection encapsuled socket
            fd (int, optional): _description_. the file decriptor for socket
            byteflow (bytes, optional): _description_. incomming message flow
        """
        pass 
        
    def _loop(self):
        pass
        
    def startloop(self):
        """This start loop goes as to start the loop for connection receiving, blocking run if 
        instance variable `is_block` is set, otherwise, runs in the thread pool.
        """
        if self.is_block:
            self._loop()
        else:
            self.threadpool.put_task(self._loop)
        
    @onexit
    def close(self, *args):
        """Exiting method for close `loop`, `epoll`, `thread pool` and `socket`
        """
        logging.debug(f"[Tcp layer] close")
        self.loop = False
        self.epctl.close()
        self.threadpool.close()
        self.sock.close()
        
        
        
@singleton
class TcpListenService(TcpService):
    """This service is a tcp listener used as a server
    """
    def __init__(self, is_block: bool, **kwargs) -> None:
        super().__init__(is_block, **kwargs)
        self.conns: Dict[int, Connection] = dict()
        self.threadpool.put_task(self._tcp_guard)
        
        
    def _rchandle(self, ops: Ops, conn: Connection, fd: int = -1, byteflow: bytes = None, *args):
        logging.debug(f"[Tcp layer] recall {ops}, {conn.addr}, {fd}, {len(byteflow) if byteflow else None}")
        if ops == Ops.Add:
            self.conns[fd] = conn
            self.epctl.register(conn.sock, select.EPOLLIN)
        elif ops == Ops.Rmv:
            try:
                conn = self.conns.get(fd)
                conn.sock.close()
            finally:
                del self.conns[fd]
            logging.debug(f"close {conn.addr}")    
        conn.last = datetime.now().timestamp()
        self.upper_rchandle(ops, conn, byteflow, *args)

    
    def _loop(self):
        ssock = self.sock
        ssock.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
        ip, port = self.kwargs.get('ip'), self.kwargs.get('port')
        ssock.bind((ip, port))
        ssock.listen(5)
        self.epctl = select.epoll()
        self.epctl.register(ssock, select.EPOLLIN)
    
        logging.info(f"start listen on ip {ip}, port {port}")
        while self.loop:
            events = self.epctl.poll()
            for fd, _ in events:
                if fd == ssock.fileno():
                    sock, addr = ssock.accept()
                    self.threadpool.put_task(self._rchandle, args=(Ops.Add, Connection(sock, addr, datetime.now().timestamp()), sock.fileno()))
                else:
                    try:
                        conn = self.conns[fd]
                        byteflow = conn.sock.recv(1024)
                        if len(byteflow) == 0:
                            raise ConnectionAbortedError()
                        self.threadpool.put_task(self._rchandle, args=(Ops.Rcv, conn, fd, byteflow))
                    except (ConnectionAbortedError, BrokenPipeError, ConnectionResetError, ConnectionRefusedError):
                        self._rchandle(Ops.Rmv, conn, fd)
                    except KeyError:
                        logging.debug(f"Conn fd {fd} not found, ignore")
    
    def startlistenloop(self):
        self.startloop()
        
        
    def _tcp_guard(self):
        max_idle = self.kwargs.get('roomgard', 3600)
        import time
        while True:
            rms = []
            for fd, conn in self.conns.items():
                if datetime.now().timestamp() - conn.last > max_idle:
                    logging.info(f"tcp guards remove {conn.addr}")
                    rms.append((conn, fd))
            for conn, fd in rms:
                self._rchandle(Ops.Rmv, conn, fd)
            time.sleep(1)
        
        
    @onexit
    def close(self, *args):
        for fd in self.conns:
            self.conns.get(fd).sock.close()
        super().close()
        
        
@singleton  
class TcpConnectService(TcpService):
    """This service is a tcp listener used as a client
    """
    def __init__(self, is_block: bool, **kwargs) -> None:
        super().__init__(is_block, **kwargs)
        
        
    def startconnectloop(self):
        self.startloop()
    
    def _loop(self):
        self.sock.connect((self.kwargs.get('ip'), self.kwargs.get('port')))
        self.epctl.register(self.sock, select.EPOLLIN)       
        self.conn.addr = self.sock.getsockname()
        while self.loop:
            events = self.epctl.poll()
            for fd, _ in events:
                if fd == self.sock.fileno():
                    try:
                        byteflow = self.sock.recv(1024)
                        if len(byteflow) == 0:
                            raise ConnectionAbortedError()
                        self.threadpool.put_task(self._rchandle, args=(Ops.Rcv, self.conn, fd, byteflow))
                    except (ConnectionAbortedError, BrokenPipeError, ConnectionResetError, ConnectionRefusedError):
                        logging.debug("server failed")
                        self.close()
                        

    def _rchandle(self, ops: Ops, conn: Connection, fd: int = -1, byteflow: bytes = None, *args):
        logging.debug(f"[Tcp layer] recall {ops}, {conn.addr}, {fd}, {len(byteflow) if byteflow else None}")
        self.upper_rchandle(ops, conn, byteflow, *args)


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-09 20:36:08
FilePath: /outlier/src/outlier/encryption/protocol.py

'''

import hashlib
import json
from datetime import datetime

KEY = "asdqwezxc"

class Encrption:    
    def entrypt(data:str, key=KEY) -> bytes:
        hashcode = int(hashlib.md5(key.encode()).hexdigest(), 16) % 128
        out = bytearray()
        for b in data.encode('ascii'):
            b ^= hashcode
            hashcode = b
            out.append(b)
        return bytes(out)  
        
    def decrypt(data:bytes, key=KEY) -> str:
        hashcode = int(hashlib.md5(key.encode()).hexdigest(), 16) % 128
        out = bytearray()
        for b in data:
            pre_b = b
            b ^= hashcode
            hashcode = pre_b
            out.append(b)
        return out.decode('ascii')
    
class Package:
    """Package for communication, encapsuling `timestamp`, `cmd` and parameters,
    better be initialized with `Package.buildpackage`
    """
    
    CMD     = "cmd"
    TIME    = "timestamp"
    
    def __init__(self, **data) -> None:
        self.data = data if data is not None else dict()

    def add_field(self, field: str, value) -> 'Package':
        self.data[field] = value
        return self
    
    def add_field_if(self, condition: bool, field: str, value) -> 'Package':
        if condition:
            self.data[field] = value
        return self
    
    def add_cmd(self, cmd: str) -> 'Package':
        self.data[Package.CMD] = cmd
        return self
    
    def encrypt(self) -> bytes:
        try:
            data = json.dumps(self.data)
            return Encrption.entrypt(data)
        except json.JSONDecodeError as e:
            # print(e)
            return b""
        
    @staticmethod
    def decrypt(byteflow: bytes) -> 'Package':
        try:
            data = Encrption.decrypt(byteflow)
            # print(data)
            data = json.loads(data)
            return Package(**data)
        except json.JSONDecodeError as e:
            # print(e)
            return Package()
    
    @staticmethod
    def buildpackage() -> 'Package':
        return Package().add_field(Package.TIME,  datetime.timestamp(datetime.now()))
    
    
    def __repr__(self) -> str:
        return f"{self.data}"  
    
    def __getitem__(a, b):
        return a.data.get(b, None)
    
    def __setitem__(a, b, c):
        a.data[b] = c
            
    def get_field(self, item, default=None):
        ret = self.data.get(item)
        return ret if ret is not None else default
    
        
    


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-10 02:19:37
FilePath: /outlier/src/outlier/server.py
'''
import logging
import time
from argparse import ArgumentParser
from datetime import datetime
from typing import List, Union

from .biz.beans import ChatMessage, Room
from .biz.bizservice import (BizRequest, BizResponse, ServerBizService, User,
                            bizserv)
from .encryption.sessionservice import ServerSessService, SessionService
from .tools.threadpool import ThreadPool
from .transmission.tcpservice import Ops, TcpListenService

from .tools.utils import gethostaddr, initlogger


class Server(ServerBizService):
    def __init__(self, sessservice: SessionService, **kwargs) -> None:
        super().__init__(sessservice, **kwargs)
        self.rooms: List[Room] = [Room("R1")]
        self.threadpool = ThreadPool()
        self.threadpool.put_task(self._room_guard)
        self.rooms[0].lastact = datetime.now().timestamp()
        
    def _findroom(self, _x: Union[str, int]) -> Union[Room, None]:
        for room in self.rooms:
            if (type(_x) == str and room.name == _x) or (type(_x) == int and room.group == _x):
                logging.debug(f"find Room {room}")
                return room
        return None
    
    def _createroom(self, name: str) -> Room:
        room = Room(name)
        self.rooms.append(room)
        return room
    
    def _findusers(self, room: Room) -> List[User]:
        return list(filter(lambda u: u.sess.group == room.group, self.users.values()))
    
    
    def _room_guard(self):
        max_idle = self.kwargs.get('roomgard', 1800)
        while True:
            rms = []
            for room in self.rooms:
                if room.lastact != -1 and datetime.now().timestamp() - room.lastact > max_idle:
                    logging.info(f"{Room} idled for {max_idle} closed")
                    for user in self._findusers(room):
                        self.leftroom(user, None)
                    rms.append(room)
            for room in rms:
                self.rooms.remove(room)        
            time.sleep(1)
        

    @bizserv()
    def chat(self, user: User, bizreq: BizRequest) -> BizResponse:
        searchroom = self._findroom(user.sess.group)
        if searchroom is not None:            
            searchroom.lastact = datetime.now().timestamp()
            searchroom.history.append(ChatMessage.parse(bizreq.param))
            return BizResponse(bizreq.param, bizreq.param, group=searchroom.group, inc=True)
        else:
            return BizResponse("error")
    
    
    @bizserv()
    def connectuser(self, user: User, bizreq: BizRequest) -> BizResponse:
        user.name = bizreq.param
        return BizResponse(f"Welcome {bizreq.param}")
    
    @bizserv()
    def infohall(self, user: User, bizreq: BizRequest) -> BizResponse:
        info = f"{Room.header()}\n" + "\n".join(map(str, self.rooms))
        return BizResponse(f"{info}", None)
    
    @bizserv()
    def enterroom(self, user: User, bizreq: BizRequest) -> BizResponse:
        name = bizreq.param
        room = self._findroom(name)
        if room is not None:
            room.lastact = datetime.now().timestamp()
            user.sess.group = room.group
            room.connects += 1
            return BizResponse(f"Enter room {room.name}", f"{user.name} entered", group=room.group, inc=False)
        else:
            room = self._createroom(name)
            room.lastact = datetime.now().timestamp()
            user.sess.group = room.group
            room.connects += 1
            return BizResponse(f"Create and enter room {room.name}", f"{user.name} entered", group=room.group, inc=False)
        
    
    @bizserv(rc=Ops.Rmv)
    def leftroom(self, user: User, bizreq: BizRequest) -> BizResponse:
        room = self._findroom(user.sess.group)
        if room is not None:
            user.sess.group = -1
            room.connects -= 1
            return BizResponse(f"leave room {room.name}", f"{user.name} left", group=room.group)
        else:
            return BizResponse()
        
    @bizserv()
    def roominfo(self, user: User, bizreq: BizRequest) -> BizResponse:
        room = self._findroom(user.sess.group)
        if room is not None:
            ret = f"Room: {room.name}\n"
            users = self._findusers(room)
            for u in users:
                ret += f"Connecting: {u.name}\tat {u.sess.conn.addr}\n"
            ret += f"room history message size: {len(room.history)}\n"
            return BizResponse(ret)
        else:
            return BizResponse('error')
        



def main():
    argparse = ArgumentParser(prog="Chat room", description="This is a chat room for your mates")
    argparse.add_argument("-l", "--log", default="INFO", type=str, choices=["DEBUG", "INFO", "ERROR", "debug", "info", "error"])
    argparse.add_argument("-lh", "--loghandler", default=None, type=str)
    argparse.add_argument("-i", "--ip",     required=False, type=str, default=gethostaddr())
    argparse.add_argument("-p", "--port",   required=False, type=int, default=8809)
        
    kwargs = vars(argparse.parse_args())
    initlogger(kwargs.get('log').upper(), filehandlename=kwargs.get('loghandler'))    

    # logger = logging.getLogger("Server")
    
    logging.info(f"Check init parameters {kwargs}")
    
    logging.info(f"Build tcp service")
    ts = TcpListenService(True, **kwargs)
    
    logging.info(f"Build Encryption service")
    ss = ServerSessService(ts, **kwargs)
    
    logging.info(f"Build server")
    bs = Server(ss, **kwargs)
    
    logging.info(f"Server start")
    bs.start()
    
if __name__ == '__main__':
    main()


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-10 02:43:49
FilePath: /outlier/src/outlier/encryption/sessionservice.py

This module is to encrypt and decrypt tcp byteflow and provide better interfaces for business layer
'''
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Set

from ..tools.decorators import singleton
from ..transmission.tcpservice import Connection, Ops, TcpService
from .protocol import Package


@dataclass
class Session:
    """Encapsuled connection and organize it with group
    """
    conn:   Connection = None
    group:  int       = -1
    
    
class SessionService:
    def __init__(self, service: TcpService, **kwargs) -> None:
        self.tsservice: TcpService = service
        self.tsservice.set_upper_rchandle(self.rchandle)
        self.upper_rchandle = lambda *args: ...
        self.kwargs = kwargs
        
        
    def set_upper_rchandle(self, upper_rchandle):
        self.upper_rchandle = upper_rchandle
        
        
    def send(self, package: Package, session: Session = None):
        try:
            byteflow = self._convert_package_to_byteflow(package)
            self.tsservice.send(byteflow, session.conn)
        except OSError as e:
            logging.debug(f"[Sess layer] send failed {session.conn.addr}")
            
    def send_group(self, bcpackage: Package, *sessions: Session):
        pass
            
        
    def rchandle(self, ops: Ops, conn: Connection, byteflow: bytes = None, *args):
        pass
        
    def _convert_byteflow_to_package(self, byteflow: bytes) -> Package:
        return Package.decrypt(byteflow) if byteflow else Package.buildpackage()
        
    def _convert_package_to_byteflow(self, package: Package) -> bytes:
        return package.encrypt() if package else Package().encrypt()
    
    
    def start(self):
        self.tsservice.startloop()
    
    def close(self, *args):
        logging.debug(f"[Sess layer] close with {args}")
        self.tsservice.close()
    
@singleton    
class ServerSessService(SessionService):
    def __init__(self, service: TcpService, **kwargs) -> None:
        super().__init__(service, **kwargs)
        self.sesss: Dict[Any, Session] = dict()
    
    def rchandle(self, ops: Ops, conn: Connection, byteflow: bytes = None, *args):
        package = self._convert_byteflow_to_package(byteflow)
        session = self.sesss.get(conn.addr)
        if ops == Ops.Add:
            session = Session(conn)
            self.sesss[conn.addr] = session
        elif ops == Ops.Rmv:
            self.sesss.pop(conn.addr)
        elif ops == Ops.Rcv:
            pass
        logging.debug(f"[Sess layer] recall {ops}, {conn.addr}, {package}")
        self.upper_rchandle(ops, session, package, *args)
    
    def send(self, package: Package, session: Session = None):
        """to send package to the certain session connecting
        Args:
            package (Package): the sending package
            session (Session, optional): the session where to
        """
        super().send(package, session)
            
                    
    def send_group(self, bcpackage: Package, *sessions: Session):
        for session in sessions:
            if session in self.sesss.values():
                self.send(bcpackage, session)
                
                
@singleton
class ConnectSessService(SessionService):
    def __init__(self, service: TcpService, **kwargs) -> None:
        super().__init__(service, **kwargs)
        self.session = Session(service.conn)
        
    def send(self, package: Package, session: Session = None):
        return super().send(package, session if session is not None else self.session)
    
    def rchandle(self, ops: Ops, conn: Connection, byteflow: bytes = None, *args):
        package = self._convert_byteflow_to_package(byteflow)
        logging.debug(f"[Sess layer] recall {ops}, {conn.addr}, {package}")
        self.upper_rchandle(ops, self.session, package, *args)
        
    


import logging
import time
from argparse import ArgumentParser
from typing import List

from .biz.beans import ChatMessage
from .biz.bizservice import BizRequest, ClientBizService, State, bizclnt
from .encryption.sessionservice import ConnectSessService, SessionService
from .transmission.tcpservice import TcpConnectService

from .tools.utils import initlogger
from .server import Server

class Client(ClientBizService):
    def __init__(self, sessservice: SessionService, **kwargs) -> None:
        super().__init__(sessservice, **kwargs)
        
        self.chatblock = False
        self.buffer: List[ChatMessage] = list()
        self.name = kwargs.get('name')
        
    def show_msg(self, package):
        print(f"{package.get_field('param')}")
        
    def show_chat(self, package):
        cm = package.get_field('param')
        cm = ChatMessage.parse(cm)
        if self.chatblock:
            self.buffer.append(cm)
        else:
            print(cm.format())
    
    def connected(self, package):
        self.atstate = State.Hall
        self.show_msg(package)
        
    def process_input(self, inputs: str):
        super().process_input(inputs)
        time.sleep(0.5)
        for cm in self.buffer:
            print(cm.format())
            
    
    @bizclnt(state=State.IDLE, bindto=Server.connectuser, recall=connected)
    def connect(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest(param=self.name)
    
    
    
    @bizclnt(state=State.Room, invokeptn="^(?!\$).+", bindto=Server.chat, recall=show_chat)
    def chat(self, inputs=None, *args, **kwargs) -> BizRequest:
        from datetime import datetime
        cm = ChatMessage(self.name, datetime.now().timestamp(), inputs)
        return BizRequest(param=cm.getattrs())
    
    
    
    @bizclnt(state=State.Hall, invoke="$info", bindto=Server.infohall, recall=show_msg, 
             desc="To glance at the details of the server and client configuartions")
    def info(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    
    @bizclnt(state=State.Hall, invoke="$exit", 
             desc="To disconnect with the server")
    def quit(self, inputs=None, *args, **kwargs):
        self.close()
        return BizRequest()
        
    
    @bizclnt(state=State.Hall, invoke="$help")
    def help(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    
    def entered(self, package):
        self.atstate = State.Room
        self.show_msg(package)
    
    @bizclnt(state=State.Hall, invoke="$room", bindto=Server.enterroom, recall=entered,
            descparams="[room name]",
            desc="To enter one of the room or create one by specified name or your username")
    def enter(self, inputs=None, *args, **kwargs) -> BizRequest:
        if len(args) == 1:
            return BizRequest(param=args[0])
        else:
            raise Exception("Parameter Error")
    
    
    
    def left(self, package):
        self.atstate = State.Hall
        self.show_msg(package)

    @bizclnt(state=State.Room, invoke="$leave", bindto=Server.leftroom, recall=left)
    def leave(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    
    @bizclnt(state=State.Room, invoke="$info", bindto=Server.roominfo, recall=show_msg,
            desc="To glance at the details of the room")
    def roominfo(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    @bizclnt(state=State.Room, invoke="$clear",
            desc="To clear the terminal")
    def clear(self, inputs=None, *args, **kwargs) -> BizRequest:
        import subprocess
        subprocess.call('reset')
        return BizRequest()
    
    
    @bizclnt(state=State.Room, invoke="$",
            desc="Switch comming message block")
    def blockswitch(self, inputs=None, *args, **kwargs) -> BizRequest:
        if self.chatblock:
            self.chatblock = False
            for cm in self.buffer:
                print(cm.format())
        else:
            self.chatblock = True
        return BizRequest()
        
        
    def start(self):
        super().start()
        time.sleep(0.5)
        self.connect()
        while True:
            a = input()
            self.process_input(a)
            
            
            



def main():
    argparse = ArgumentParser(prog="Chat room", description="This is a chat room for your mates")
    argparse.add_argument("-l", "--log", default="INFO", type=str, choices=["DEBUG", "INFO", "ERROR", "debug", "info", "error"])
    argparse.add_argument("-n", "--name", required=True, type=str)
    argparse.add_argument("-i", "--ip",   required=True, type=str)
    argparse.add_argument("-p", "--port", required=False, type=int, default=8809)
        
    kwargs = vars(argparse.parse_args())
    initlogger(kwargs.get('log').upper(), filehandlename=kwargs.get('loghandler'))    
    
    ts = TcpConnectService(False, **kwargs)
    ss = ConnectSessService(ts, **kwargs)
    bs = Client(ss, **kwargs)
    bs.start()  

    
if __name__ == '__main__':
    main()



'''
Date: 2023-03-10 02:53:02
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-10 02:53:06
FilePath: /outlier/setup.py
'''
import setuptools

with open("README.md", 'r') as f:
    long_description = f.read()
    
setuptools.setup(
    name="outlierchat",
    version="0.0.9",
    author="Thetis",
    author_email="736396627@qq.com",
    description="A simple server to build a secret channel for chat.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    classifers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache License",
        "Operating System :: OS Independent"
    ],
    python_requires='>3.6'
)


import sys
import time
import unittest
from pathlib import Path
from tools.threadpool import ThreadPool

sys.path.append(Path(__file__).parent.parent / "src" / "outlier")

class testThreadPool(unittest.TestCase):
    def test_norm(self):
        starttime = time.time()
        t = ThreadPool(10, False)
        t.put_task(ThreadPool.template_loop, args=(1,))
        t.put_task(ThreadPool.template_loop, args=(5,))
        t.put_task(ThreadPool.template_loop, args=(2,))
        t.put_task(ThreadPool.template_loop, args=(4,))
        t.put_task(ThreadPool.template_loop, args=(3,))
        t.queue.join()
        endtime = time.time()
        self.assertAlmostEqual(endtime - starttime, 5, delta=0.1)
        
    def test_thread_dead(self):
        starttime = time.time()
        t = ThreadPool(2, False)
        t.put_task(lambda x: 1/ x, args=(0,))
        t.put_task(ThreadPool.template_loop, args=(5,))
        time.sleep(2)
        t.put_task(ThreadPool.template_loop, args=(2,))
        t.queue.join()
        endtime = time.time()
        self.assertAlmostEqual(endtime - starttime, 5, delta=0.1)



__all__ = ['server', 'client', 'biz', 'encryption' ,'tools', 'transmission']

from . import server
from . import client
from . import biz
from . import encryption
from . import tools
from . import transmission